@@ -2,7 +2,8 @@ import cv2
 import numpy as np
 
 
-def get_default_opts():
+def get_default_opts(model="pinhole"):
+
     default_opts = {
         'debug': True,  # Will enable advanced tests and outputs
         'coord_cam': 0,  # Reference camera that defines the multicam coordinate system
@@ -15,23 +16,19 @@ def get_default_opts():
         # Do not perform detection and single cam calibration. (Disable mostly for development.)
         'numerical_jacobian': False,  # Use 2-point numerical jacobian instead of jax.jacobian
         'optimize_board_poses': False,  # Optimize individual board poses then all params again. In a test,
-                                        #  optimality was already reached after a first general optimization
-        'free_vars': {
-            'cam_pose': True,
-            'board_poses': True,
-            'A': np.asarray([[True, False, True],  # a   c   u   (c is skew and should not be necessary)
-                             [False, True, True],  # 0   b   v
-                             [False, False, False],  # 0   0   1
-                             ]),
-            'k': np.asarray([1, 1, -1, -1, -1]),  # 1: optimize, 0: leave const, -1: force 0
-        },
+        #  optimality was already reached after a first general optimization
+
+        'max_allowed_res': 1.0,  # In pixels. Reject the pose with higher error and insert 'nearby' pose with lower
+        # error while optimizing individual board poses.
+
+        'free_vars': get_free_vars(model),
         'detection': {
-            'inter_frame_dist': 0.0,
+            'inter_frame_dist': 0.0,  # In pixels
             'aruco_calibration': {
-                'flags': (cv2.CALIB_ZERO_TANGENT_DIST + cv2.CALIB_FIX_K3),
+                'flags': get_flags(model),
                 'criteria': (cv2.TermCriteria_COUNT + cv2.TermCriteria_EPS,
                              30,
-                             1.1920928955078125e-07),  # float(np.finfo(np.float32).eps)
+                             float(np.finfo(np.float32).eps)),
             },
             'aruco_detect': {
                 'parameters': get_detector_parameters_opts(),
@@ -53,8 +50,8 @@ def get_default_opts():
             'gtol': 1e-8,
             'x_scale': 'jac',
             'loss': 'linear',
-            'tr_solver': 'exact',
-            'max_nfev': 100,
+            'tr_solver': 'lsmr',
+            'max_nfev': 150,
             'verbose': 2,
         },
     }
@@ -62,6 +59,39 @@ def get_default_opts():
     return default_opts
 
 
+def get_free_vars(model: str):
+    free_vars = {
+        'cam_pose': True,
+        'board_poses': True,
+        'A': np.asarray([[True, False, True],  # a   c   u   (c is skew and should not be necessary)
+                         [False, True, True],  # 0   b   v
+                         [False, False, False],  # 0   0   1
+                         ]),
+        'xi': False,
+        'k': np.asarray([1, 1, -1, -1, -1]),  # 1: optimize, 0: leave const, -1: force 0
+    }
+
+    if model == "omnidir":
+        # 'A' or 'K' (opencv-omnidir notation) - camera matrix
+        # 'k' or 'D' (opencv-omnidir notation) - distortion coeffs
+        free_vars['xi'] = True
+        free_vars['k'] = np.asarray([1, 1, 1, 1, 1])
+
+    return free_vars
+
+
+def get_flags(model: str):
+    flags = cv2.CALIB_ZERO_TANGENT_DIST + cv2.CALIB_FIX_K3
+
+    if model == "omnidir":
+        # "omnidir"
+        # Should be set to None or 0, if no flags are to be used. Use 0, None is causing error with scipy_io_savemat
+        # return (cv2.omnidir.CALIB_FIX_P1 + cv2.omnidir.CALIB_FIX_P2)
+        flags = cv2.omnidir.CALIB_FIX_SKEW
+
+    return flags
+
+
 def get_detector_parameters_opts():
     detector_parameters = {  # SPOT for detector params
         'cornerRefinementMethod': cv2.aruco.CORNER_REFINE_SUBPIX,
