@@ -20,9 +20,11 @@ from calibcam import helper, camfunctions, board, compatibility
 
 from calibcam.calibrator_opts import get_default_opts
 from calibcam.pose_estimation import estimate_cam_poses
+from calibcam.single_camcalibration import calibrate_single_camera
+
 
 class CamCalibrator:
-    def __init__(self, recordings, board_name=None, data_path=None, opts=None):
+    def __init__(self, recordings, board_name=None, data_path=None, calibs_init=None, opts=None):
         if opts is None:
             opts = {}
 
@@ -43,9 +45,15 @@ class CamCalibrator:
 
         # Options
         self.opts = {}
-        self.load_opts(opts, self.data_path)
+        self.opts = self.load_opts(opts, self.data_path)
+
+        # Recordings
         self.load_recordings(recordings)
 
+        # Calibs initialization
+        self.calibs_init = None
+        self.calibs_init = self.load_calibs_init(calibs_init, self.data_path)
+
         return
 
     def get_board_params_from_name(self, board_name):
@@ -55,12 +63,22 @@ class CamCalibrator:
             board_params = board.get_board_params(Path(self.rec_file_names[0]).parent)
         return board_params
 
-    def load_opts(self, opts, data_path=None):
+    @staticmethod
+    def load_calibs_init(calibs_init, data_path=None):
+        calib_init_path = data_path + "/multicam_calibration.npy"
+        if calibs_init is None and data_path is not None and os.path.isfile(calib_init_path):
+            print(f"Loading initialization from {calib_init_path}")
+            calibs_init = np.load(calib_init_path, allow_pickle=True).item()["calibs"]
+
+        return calibs_init
+
+    @staticmethod
+    def load_opts(opts, data_path=None):
         if data_path is not None and os.path.isfile(data_path + "/opts.npy"):
             fileopts = np.load(data_path + "/opts.npy", allow_pickle=True).item()
             opts = helper.deepmerge_dicts(opts, fileopts)
 
-        self.opts = helper.deepmerge_dicts(opts, get_default_opts())
+        return helper.deepmerge_dicts(opts, get_default_opts())
 
     def load_recordings(self, recordings):
         # check if input files are valid files
@@ -97,17 +115,16 @@ class CamCalibrator:
     def perform_multi_calibration(self):
         n_corners = (self.board_params["boardWidth"] - 1) * (self.board_params["boardHeight"] - 1)
         required_corner_idxs = [0,
-                            self.board_params["boardWidth"] - 1,
-                            (self.board_params["boardWidth"] - 1) * (self.board_params["boardHeight"] - 2),
-                            (self.board_params["boardWidth"] - 1) * (self.board_params["boardHeight"] - 1) - 1,
-                            ]  # Corners that we require to be detected for pose estimation
+                                self.board_params["boardWidth"] - 2,
+                                (self.board_params["boardWidth"] - 1) * (self.board_params["boardHeight"] - 2),
+                                (self.board_params["boardWidth"] - 1) * (self.board_params["boardHeight"] - 1) - 1,
+                                ]  # Corners that we require to be detected for pose estimation
         if self.opts["optimize_only"]:  # We expect that detections and single cam calibs are already present
             preoptim = np.load(self.data_path + '/preoptim.npy', allow_pickle=True)[()]
             preoptim = compatibility.update_preoptim(preoptim, n_corners)
 
             calibs_single = preoptim['info']['other']['calibs_single']
-            if 'used_frames_ids' in preoptim['info']:
-                used_frames_ids = preoptim['info']['used_frames_ids']
+            used_frames_ids = preoptim['info']['used_frames_ids']
             corners = preoptim['info']['corners']
 
             # We just redo this since it is fast and the output may help
@@ -124,13 +141,18 @@ class CamCalibrator:
             corners, used_frames_ids = \
                 detect_corners(self.rec_file_names, self.n_frames, self.board_params, self.opts)
 
-            # perform single calibration
-            calibs_single = self.perform_single_cam_calibrations(corners)
+            if 'internals' in self.opts:
+                calibs_single = self.perform_board_position_estimation(self.opts["internals"], corners)
+
+            else:
+                #  perform single calibration
+                calibs_single = self.perform_single_cam_calibrations(corners, calibs_init=self.calibs_init)
 
             # analytically estimate initial camera poses
             calibs_multi = estimate_cam_poses(calibs_single, self.opts, corners=corners,
                                               required_corner_idxs=required_corner_idxs)
 
+
             # Save intermediate result, for dev purposes on optimization (quote code above and unquote code below)
             # pose_params = optimization.make_common_pose_params(calibs_multi, corners)
             result = self.build_result(calibs_multi,
@@ -141,7 +163,8 @@ class CamCalibrator:
 
         if self.opts['debug']:
             args, vars_free = make_optim_input(self.board_params, calibs_multi, corners, self.opts)
-            test_objective_function(calibs_multi, vars_free, args, corners, self.board_params, individual_poses=True)
+            test_objective_function(calibs_multi, vars_free, args, corners, self.board_params,
+                                    individual_poses=True)
 
         print('OPTIMIZING ALL POSES')
         # self.plot(calibs_single, corners, used_frames_ids, self.board_params, 3, 35)
@@ -149,7 +172,8 @@ class CamCalibrator:
 
         if self.opts['debug']:
             calibs_fit = helper.combine_calib_with_board_params(calibs_fit, rvecs_boards, tvecs_boards)
-            test_objective_function(calibs_fit, min_result.x, args, corners, self.board_params, individual_poses=True)
+            test_objective_function(calibs_fit, min_result.x, args, corners, self.board_params,
+                                    individual_poses=True)
 
         print('OPTIMIZING ALL PARAMETERS I')
         calibs_fit, rvecs_boards, tvecs_boards, min_result, args = self.optimize_calibration(corners, calibs_fit)
@@ -159,22 +183,26 @@ class CamCalibrator:
         if self.opts["optimize_board_poses"]:
             if self.opts['debug']:
                 calibs_fit = helper.combine_calib_with_board_params(calibs_fit, rvecs_boards, tvecs_boards)
-                test_objective_function(calibs_fit, min_result.x, args, corners, self.board_params, individual_poses=True)
+                test_objective_function(calibs_fit, min_result.x, args, corners, self.board_params,
+                                        individual_poses=True)
 
             print('OPTIMIZING BOARD POSES')
-            calibs_fit, rvecs_boards, tvecs_boards, min_result, args = self.optimize_board_poses(corners, calibs_fit)
+            calibs_fit, rvecs_boards, tvecs_boards, min_result, args = self.optimize_board_poses(corners, calibs_fit,
+                                                                                                 prev_fun=min_result.fun)
             calibs_fit = helper.combine_calib_with_board_params(calibs_fit, rvecs_boards, tvecs_boards)
 
             if self.opts['debug']:
                 args, vars_free = make_optim_input(self.board_params, calibs_fit, corners, self.opts)
-                test_objective_function(calibs_fit, vars_free, args, corners, self.board_params, individual_poses=True)
+                test_objective_function(calibs_fit, vars_free, args, corners, self.board_params,
+                                        individual_poses=True)
 
             print('OPTIMIZING ALL PARAMETERS II')
             calibs_fit, rvecs_boards, tvecs_boards, min_result, args = self.optimize_calibration(corners, calibs_fit)
             # No board poses in final calibration!
 
         calibs_test = helper.combine_calib_with_board_params(calibs_fit, rvecs_boards, tvecs_boards, copy=True)
-        test_objective_function(calibs_test, min_result.x, args, corners, self.board_params, individual_poses=True)
+        test_objective_function(calibs_test, min_result.x, args, corners, self.board_params,
+                                individual_poses=True)
 
         result = self.build_result(calibs_fit,
                                    corners=corners, used_frames_ids=used_frames_ids,
@@ -191,55 +219,58 @@ class CamCalibrator:
         print('FINISHED MULTI CAMERA CALIBRATION')
         return
 
-    @staticmethod
-    def calibrate_single_camera(corners_cam, sensor_size, board_params, opts, mask=None):
+    def perform_board_position_estimation(self, calibs_single, corners):
+        print('ESTIMATE BOARD POSITIONS')
+
+        for i_cam, calib in enumerate(calibs_single):
+            calibs_single[i_cam] = self.estimate_board_positions_in_single_cam(calib, corners[i_cam])
+
+        return calibs_single
+
+    def estimate_board_positions_in_single_cam(self, calib, corners_cam, mask=None):
         if mask is None:
-            mask = np.sum(~np.isnan(corners_cam[:, :, 1]), axis=1) > 0  # Test for degeneration should be performed beforehand and respective frames excluded from corner array
-
-        n_used_frames = np.sum(mask)
-
-        if n_used_frames == 0:
-            return []
+            mask = np.sum(~np.isnan(corners_cam[:, :, 1]), axis=1) > 0
 
         corners_nn = corners_cam[mask]
         corners_use, ids_use = helper.corners_array_to_ragged(corners_nn)
 
-        cal_res = cv2.aruco.calibrateCameraCharucoExtended(corners_use,  # noqa
-                                                           ids_use,
-                                                           board.make_board(board_params),
-                                                           sensor_size,
-                                                           None,
-                                                           None,
-                                                           **opts['detection']['aruco_calibration'])
+        board_positions = Parallel(n_jobs=int(np.floor(multiprocessing.cpu_count())))(
+            delayed(self.estimate_single_board_position)(calib,
+                                                         corners_use[i_pose],
+                                                         ids_use[i_pose],
+                                                         self.board_params,
+                                                         self.opts)
+            for i_pose in range(len(corners_use)))
 
-        rvecs = np.empty(shape=(mask.size, 3))
-        rvecs[:] = np.NaN
-        tvecs = np.empty(shape=(mask.size, 3))
-        tvecs[:] = np.NaN
-        retval, A, k,  = cal_res[0:3]
+        calib['rvecs'] = np.full((corners_cam.shape[0], 3), np.nan)
+        calib['tvecs'] = np.full((corners_cam.shape[0], 3), np.nan)
+        calib['frames_mask'] = mask
 
-        rvecs[mask, :] = np.asarray(cal_res[3])[..., 0]
-        tvecs[mask, :] = np.asarray(cal_res[4])[..., 0]
+        pose_idxs = np.where(mask)[0]
+        for pose_idx, pos in zip(pose_idxs, board_positions):
+            if pos[0]:
+                calib['rvecs'][pose_idx] = pos[1][:,0]
+                calib['tvecs'][pose_idx] = pos[2][:,0]
 
-        cal = {
-            'rvec_cam': np.asarray([0., 0., 0.]),
-            'tvec_cam': np.asarray([0., 0., 0.]),
-            'A': np.asarray(A),
-            'k': np.asarray(k).ravel(),
-            'rvecs': np.asarray(rvecs),
-            'tvecs': np.asarray(tvecs),
-            'repro_error': retval,  # Not that from here on values are NOT expanded to full frames range, see frames_mask
-            'std_intrinsics': cal_res[5],
-            'std_extrinsics': cal_res[6],
-            'per_view_errors': cal_res[7],
-            'frames_mask': mask,
-        }
-        print('Finished single camera calibration.')
-        return cal
+        return calib
 
-    def perform_single_cam_calibrations(self, corners):
+    @staticmethod
+    def estimate_single_board_position(calib, corners, ids, board_params, opts):
+        if len(ids)<4:
+            return 0, np.full((3,), np.nan), np.full((3,), np.nan)
+
+        retval, rvec, tvec = cv2.solvePnP(board.make_board_points(board_params)[ids].reshape((-1,3)),
+                                          corners.reshape((-1, 2)),
+                                          calib["A"], calib["k"],
+                                          flags=cv2.SOLVEPNP_IPPE)
+        return retval, rvec, tvec
+
+    def perform_single_cam_calibrations(self, corners, calibs_init=None):
         print('PERFORM SINGLE CAMERA CALIBRATION')
 
+        if calibs_init is None:
+            calibs_init = [None for i in corners]
+
         # calibs_single = [self.calibrate_single_camera(corners[i_cam],
         #                                               camfunctions.get_header_from_reader(self.readers[i_cam])['sensorsize'],
         #                                               self.board_params,
@@ -247,11 +278,12 @@ class CamCalibrator:
         #                  for i_cam in range(len(self.readers))]
         print(int(np.floor(multiprocessing.cpu_count())))
         calibs_single = Parallel(n_jobs=int(np.floor(multiprocessing.cpu_count())))(
-            delayed(self.calibrate_single_camera)(corners[i_cam],
-                                                  camfunctions.get_header_from_reader(self.readers[i_cam])[
-                                                      'sensorsize'],
-                                                  self.board_params,
-                                                  self.opts)
+            delayed(calibrate_single_camera)(corners[i_cam],
+                                             camfunctions.get_header_from_reader(self.readers[i_cam])[
+                                                 'sensorsize'],
+                                             self.board_params,
+                                             self.opts,
+                                             calib_init=calibs_init[i_cam])
             for i_cam in range(len(self.readers)))
 
         for i_cam, calib in enumerate(calibs_single):
@@ -273,13 +305,14 @@ class CamCalibrator:
         pose_opts = deepcopy(opts)
         pose_opts['free_vars']['A'][:] = False
         pose_opts['free_vars']['k'][:] = False
+        pose_opts['free_vars']['xi'] = False
 
         calibs_fit, rvecs_boards, tvecs_boards, min_result, args = \
             camfunctions.optimize_calib_parameters(corners, calibs_multi, board_params, opts=pose_opts)
 
         return calibs_fit, rvecs_boards, tvecs_boards, min_result, args
 
-    def optimize_board_poses(self, corners, calibs_multi, opts=None, board_params=None):
+    def optimize_board_poses(self, corners, calibs_multi, opts=None, board_params=None, prev_fun=None):
         if opts is None:
             opts = self.opts
         if board_params is None:
@@ -289,20 +322,33 @@ class CamCalibrator:
         pose_opts['free_vars']['cam_pose'] = False
         pose_opts['free_vars']['A'][:] = False
         pose_opts['free_vars']['k'][:] = False
+        pose_opts['free_vars']['xi'] = False
 
         calibs_multi_pose = deepcopy(calibs_multi)
         rvecs_boards = calibs_multi[0]["rvecs"]
         tvecs_boards = calibs_multi[0]["tvecs"]
+
+        if prev_fun is not None:
+            prev_fun = prev_fun.reshape(corners.shape)
+            good_poses = set(np.arange(prev_fun.shape[1]))
+            for i_cam in range(prev_fun.shape[0]):
+                good_poses = good_poses - set(np.where(prev_fun[i_cam] > opts['max_allowed_res'])[0])
+            good_poses = list(good_poses)
+        else:
+            good_poses = list(range(len(rvecs_boards)))
+
         print(f"Optimizing {len(calibs_multi[0]['rvecs'])} poses: ", end='')
         for i_pose in range(len(calibs_multi[0]["rvecs"])):
             print(".", end='', flush=True)
             corners_pose = corners[:, [i_pose]]
             for calib, calib_orig in zip(calibs_multi_pose, calibs_multi):
-                calib["rvecs"] = calib_orig["rvecs"][[i_pose]]
-                calib["tvecs"] = calib_orig["tvecs"][[i_pose]]
+                nearest_i_pose = helper.nearest_element(i_pose, good_poses)  # nearest_i_pose = i_pose if i_pose in good_poses
+                calib["rvecs"] = calib_orig["rvecs"][[nearest_i_pose]]
+                calib["tvecs"] = calib_orig["tvecs"][[nearest_i_pose]]
 
             calibs_fit_pose, rvecs_boards[i_pose], tvecs_boards[i_pose], min_result, args = \
-                camfunctions.optimize_calib_parameters(corners_pose, calibs_multi_pose, board_params, opts=pose_opts, verbose=0)
+                camfunctions.optimize_calib_parameters(corners_pose, calibs_multi_pose, board_params, opts=pose_opts,
+                                                       verbose=0)
         print()
         return calibs_fit_pose, rvecs_boards, tvecs_boards, min_result, args
 
@@ -355,6 +401,7 @@ class CamCalibrator:
 
         # savemat cannot deal with none!
         if min_result is not None:
+            result['info']['fun_final'] = min_result.fun
             result['info']['cost_val_final'] = min_result.cost
             result['info']['optimality_final'] = min_result.optimality
 
@@ -387,7 +434,10 @@ class CamCalibrator:
         plt.imshow(cv2.aruco.drawDetectedCornersCharuco(im, corners_use[fidx], ids_use[fidx]))
 
         board_coords_3d = R.from_rotvec(r).apply(board_coords_3d_0) + t
-        board_coords_3d = camfunctions_ag.board_to_ideal_plane(board_coords_3d)
+
+        board_coords_3d = camfunctions_ag.board_to_unit_sphere(board_coords_3d)
+        board_coords_3d = camfunctions_ag.shift_camera(board_coords_3d, calibs[cidx]['xi'].squeeze()[0])
+        board_coords_3d = camfunctions_ag.to_ideal_plane(board_coords_3d)
 
         board_coords_3d_nd = camfunctions_ag.ideal_to_sensor(board_coords_3d, calibs[cidx]['A'])
 
